#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>       /* printk() */
#include <linux/slab.h>         /* kmalloc() */
#include <linux/fs.h>           /* everything... */
#include <linux/errno.h>        /* error codes */
#include <linux/types.h>        /* size_t */
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/hdreg.h>        /* HDIO_GETGEO */
#include <asm/system.h>         /* cli(), *_flags */
#include <linux/ioctl.h>
#include <linux/vmalloc.h>
#include <linux/ts_sbc.h>
#include <asm/uaccess.h>
#include <asm/io.h>

#define MAJOR_NR sdcard_major   /* force definitions on in blk.h */
int sdcard_major;               /* must be declared before including blk.h */

#include <linux/blk.h>
#include <linux/blkpg.h>        /* blk_ioctl() */
#include <linux/sched.h>

extern void __udelay(unsigned int);
#define udelay(x) __udelay((x))
#include "sd.c"

MODULE_AUTHOR("Technologic Systems");
MODULE_LICENSE("Proprietary");
MODULE_PARM(io, "l");
MODULE_PARM(name, "s");

static unsigned char *name = "sdcard0";
static unsigned int io = 0x13000000;
static int sdcard_size;
static int sdcard_sizes[16];
static struct hd_struct sdcard_partitions[16];
static devfs_handle_t devfs_handle;
static struct gendisk sdcard_gendisk;
static unsigned char *start;

/* Forwards */
static int sdcard_revalidate(kdev_t i_rdev);
static struct block_device_operations sdcard_bdops;


static int sdcard_open(struct inode *inode, struct file *filp)
{
    MOD_INC_USE_COUNT;
    return 0;
}


static int sdcard_release(struct inode *inode, struct file *filp)
{
    MOD_DEC_USE_COUNT;
    return 0;
}


static int sdcard_ioctl(struct inode *inode, struct file *filp,
                        unsigned int cmd, unsigned long arg)
{
    int err;
    struct hd_geometry geo;

    switch (cmd) {
    case HDIO_GETGEO:
        /*
         * get geometry: we have to fake one...  trim the size to a
         * multiple of 512 (256k): tell we have 32 sectors, 16 heads,
         * whatever cylinders.
         */

        // TODO: This should match what the SD cards look like in a 
        // USB adapter dongle.  -- JO

        err = !access_ok(VERIFY_WRITE, arg, sizeof(geo));
        if (err)
            return -EFAULT;

        geo.cylinders = sdcard_size / (16 * 32);
        geo.heads = 16;
        geo.sectors = 32;
        geo.start = 0;

        if (copy_to_user((void *) arg, &geo, sizeof(geo)))
            return -EFAULT;

        return 0;

    default:
        /*
         * For ioctls we don't understand, let the block layer handle them.
         */

        return blk_ioctl(inode->i_rdev, cmd, arg);
    }

    return -ENOTTY;
}


static int sdcard_check_change(kdev_t i_rdev)
{
    if (sdping()) {
    	sdinit(start);
	return 1;
    } else return 0;
}


static struct block_device_operations sdcard_bdops = {
  open:			sdcard_open,
  release:		sdcard_release,
  ioctl:		sdcard_ioctl,
  revalidate:		sdcard_revalidate,
  check_media_change:	sdcard_check_change,
};


static int sdcard_revalidate(kdev_t i_rdev)
{

    /* first clear old partition information */
    memset(sdcard_gendisk.sizes + 1, 0, 15 * sizeof(int));
    memset(sdcard_gendisk.part + 1, 0, 15 * sizeof(struct hd_struct));

    sdcard_size = sdcard_gendisk.part[0].nr_sects = sdsize();

    /* then fill new info */
    register_disk(&sdcard_gendisk, i_rdev, 16, &sdcard_bdops, sdcard_size);

    return 0;
}


static int sdcard_make_request(request_queue_t * q, int rw,
                               struct buffer_head *bh)
{
    unsigned int ret = 0;
    bh->b_rsector += sdcard_partitions[MINOR(bh->b_rdev)].start_sect;

    switch (rw) {
    case READ:
        ret = sdread(bh->b_rsector, bh->b_data, bh->b_size / 512);
	if (ret) {
	    sdinit(start);
            ret = sdread(bh->b_rsector, bh->b_data, bh->b_size / 512);
	}
        bh->b_end_io(bh, !ret);
        break;
    case WRITE:
        refile_buffer(bh);
        ret = sdwrite(bh->b_rsector, bh->b_data, bh->b_size / 512);
	if (ret && !sdwprot()) {
  	    sdinit(start);
       	    ret = sdwrite(bh->b_rsector, bh->b_data, bh->b_size / 512);
	}
	if (ret == 0) mark_buffer_uptodate(bh, 1);
        bh->b_end_io(bh, !ret);
        break;
    default:
        bh->b_end_io(bh, 0);
        break;
    }
    return 0;
}


static int sdcard_init(void)
{
    int model = inw(TS7XXX_MODEL) & 0x7;
    int pldrev = inw(TS7XXX_PLDREV) & 0x7;
    int opts = inw(0x02400000 + TS7XXX_IO16_BASE) & 0x7;
    unsigned int scratchreg0 = inl(SYSCON_SCRREG0); 

    start = (unsigned char *) __ioremap(io, 4, 0);

    printk("%s: Technologic Systems SD card controller, address 0x%x\n",
           name, io);

    /*
    if (!((model == TS7260 && pldrev >= 5 && (opts & 0x4)) || model == TS7300)) {
    	printk("%s: no SD card option present on this board\n", name);
	return 1;
    }
    */

    sdcard_size = 0;
    if (scratchreg0) {
    	sdcard_size = sdfastinit(start, scratchreg0);
	outl(0, SYSCON_SCRREG0);
    }
    if (sdcard_size == 0) sdcard_size = sdinit(start);

    if (sdcard_size == 0) {
    	printk("%s: no card found\n", name);
	return 1;
    } else {
    	printk("%s: card size %d sectors\n", name, sdcard_size);
    }

    sdcard_major = register_blkdev(0, name, &sdcard_bdops);
    devfs_handle = devfs_mk_dir(NULL, name, NULL);

    memset(sdcard_sizes, 0, 16 * sizeof(int));
    sdcard_sizes[0] = sdcard_size;

    memset(sdcard_partitions, 0, 16 * sizeof(struct hd_struct));
    sdcard_partitions[0].nr_sects = sdcard_size;

    sdcard_gendisk.minor_shift = 4;
    sdcard_gendisk.max_p = 16;
    sdcard_gendisk.fops = &sdcard_bdops;
    sdcard_gendisk.major = sdcard_major;
    sdcard_gendisk.major_name = name;
    sdcard_gendisk.part = sdcard_partitions;
    sdcard_gendisk.nr_real = 1;
    sdcard_gendisk.sizes = sdcard_sizes;

    add_gendisk(&sdcard_gendisk);

    blk_size[sdcard_major] = sdcard_sizes;
    hardsect_size[sdcard_major] = NULL;
    blksize_size[sdcard_major] = NULL;
    read_ahead[sdcard_major] = 0;

    blk_queue_make_request(BLK_DEFAULT_QUEUE(sdcard_major),
                           sdcard_make_request);
    register_disk(&sdcard_gendisk, MKDEV(sdcard_major, 0), 16,
                  &sdcard_bdops, sdcard_size);

    EXPORT_NO_SYMBOLS;

    return 0;
}


static void sdcard_cleanup(void)
{
    unregister_blkdev(sdcard_major, name);
    devfs_unregister(devfs_handle);

    blk_size[sdcard_major] = NULL;

    del_gendisk(&sdcard_gendisk);
    __iounmap((unsigned char *) io);
}


int init_module(void)
{
    return sdcard_init();
}


void cleanup_module(void)
{
    return sdcard_cleanup();
}
